@@ -1,17 +1,15 @@
-import sys
 import asyncio
 import asyncio.events as events
-import threading
+import sys
+from heapq import heappop
 
 
 def apply(loop=None):
-    """
-    Patch asyncio to make its event loop reentrent.
-    """
+    """Patch asyncio to make its event loop reentrent."""
     loop = loop or asyncio.get_event_loop()
     if not isinstance(loop, asyncio.BaseEventLoop):
         raise ValueError('Can\'t patch loop of type %s' % type(loop))
-    if hasattr(loop, '_run_forever_orig'):
+    if getattr(loop, '_nest_patched', None):
         # already patched
         return
     _patch_asyncio()
@@ -44,57 +42,60 @@ def _patch_asyncio():
 
 
 def _patch_loop(loop):
-    """
-    Patch loop to make it reentrent.
-    """
-    if sys.version_info >= (3, 7, 0):
-        set_coro_tracking = loop._set_coroutine_origin_tracking
-    else:
-        set_coro_tracking = loop._set_coroutine_wrapper
+    """Patch loop to make it reentrent."""
 
-    def run_forever(self):
+    def run_until_complete(self, future):
         self._check_closed()
-        old_thread_id = self._thread_id
-        old_running_loop = events._get_running_loop()
-        set_coro_tracking(self._debug)
-        self._thread_id = threading.get_ident()
+        events._set_running_loop(self)
+        f = asyncio.ensure_future(future, loop=self)
+        if f is not future:
+            f._log_destroy_pending = False
+        while not f.done():
+            self._run_once()
+            if self._stopping:
+                break
+        if not f.done():
+            raise RuntimeError('Event loop stopped before Future completed.')
+        return f.result()
 
-        if self._asyncgens is not None:
-            old_agen_hooks = sys.get_asyncgen_hooks()
-            sys.set_asyncgen_hooks(
-                firstiter=self._asyncgen_firstiter_hook,
-                finalizer=self._asyncgen_finalizer_hook)
-        try:
-            events._set_running_loop(self)
-            while True:
-                try:
-                    self._run_once()
-                except IndexError:
-                    # Ignore 'pop from an empty deque' errors.
-                    # This happens when all ready handles have already been
-                    # processed but _run_once expects there to be more.
-                    # Since the handles have been processed anyway it is
-                    # safe to ignore this.
-                    pass
-                if self._stopping:
-                    break
-        finally:
-            self._stopping = False
-            self._thread_id = old_thread_id
-            events._set_running_loop(old_running_loop)
-            set_coro_tracking(False)
-            if self._asyncgens is not None:
-                sys.set_asyncgen_hooks(*old_agen_hooks)
+    def _run_once(self):
+        """
+        Simplified re-implementation of asyncio's _run_once that
+        runs handles as they become ready.
+        """
+        now = self.time()
+        ready = self._ready
+        scheduled = self._scheduled
+        while scheduled and scheduled[0]._cancelled:
+            heappop(scheduled)
+
+        timeout = 0 if ready or self._stopping \
+            else min(max(0, scheduled[0]._when - now), 10) if scheduled \
+            else None
+        event_list = self._selector.select(timeout)
+        self._process_events(event_list)
+
+        while scheduled and scheduled[0]._when < now + self._clock_resolution:
+            handle = heappop(scheduled)
+            ready.append(handle)
+
+        while ready:
+            handle = ready.popleft()
+            if not handle._cancelled:
+                handle._run()
+        handle = None
 
     cls = loop.__class__
-    cls._run_forever_orig = cls.run_forever
-    cls.run_forever = run_forever
+    cls._run_once_orig = cls._run_once
+    cls._run_once = _run_once
+    cls._run_until_complete_orig = cls.run_until_complete
+    cls.run_until_complete = run_until_complete
+    cls._nest_patched = True
 
 
 def _patch_task():
-    """
-    Patch the Task's step and enter/leave methods to make it reentrant.
-    """
+    """Patch the Task's step and enter/leave methods to make it reentrant."""
+
     def step(task, exc=None):
         curr_task = curr_tasks.get(task._loop)
         try:
@@ -126,13 +127,23 @@ def _patch_task():
 
 
 def _patch_handle():
-    """
-    Patch Handle to allow recursive calls.
-    """
+    """Patch Handle to allow recursive calls."""
+
+    def update_from_context(ctx):
+        """Copy context ctx to currently active context."""
+        for var in ctx:
+            var.set(ctx[var])
+
     def run(self):
+        """
+        Run the callback in a sub-context, then copy any sub-context vars
+        over to the Handle's context.
+        """
         try:
             ctx = self._context.copy()
             ctx.run(self._callback, *self._args)
+            if ctx:
+                self._context.run(update_from_context, ctx)
         except Exception as exc:
             cb = format_helpers._format_callback_source(
                 self._callback, self._args)
@@ -149,4 +160,4 @@ def _patch_handle():
 
     if sys.version_info >= (3, 7, 0):
         from asyncio import format_helpers
-        asyncio.events.Handle._run = run
+        events.Handle._run = run
